/* 
   divx4file_core.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "avilib.h"
#include "encore2.h"
#include "lame3.70/lame.h"
#include "../NVrec.h"

#define MP3BUFSIZE 65536

struct divx4file_core_t {
	avi_t * avi;
	void * dh;
	lame_global_flags lgf;
	unsigned char * cbuf;
	unsigned char mbuf[MP3BUFSIZE];
	ENC_FRAME enc_frame;
	ENC_RESULT enc_result;
	int ashift;
};

struct file_core_t divx4file_core;

void divx4file_core_audio(struct file_core_t * fct, unsigned char *buf, int n)
{
	struct divx4file_core_t * divx4c = (struct divx4file_core_t *)fct->priv;
	int i;
	
	if(!divx4c)
		return;
	i = lame_encode_buffer_interleaved(&divx4c->lgf, (short int *)buf, n >> divx4c->ashift, divx4c->mbuf, MP3BUFSIZE);
	if(i > 0)
		AVI_write_audio(divx4c->avi, divx4c->mbuf, i);
}

void divx4file_core_close(struct file_core_t * fct)
{
	struct divx4file_core_t * divx4c = (struct divx4file_core_t *)fct->priv;
	int i;
	
	if(!divx4c)
		return;
	if(encore(divx4c->dh, ENC_OPT_RELEASE, NULL, NULL) != ENC_OK)
		fprintf(stderr, "Error freeing encore device!\n");
	i = lame_encode_finish(&divx4c->lgf, divx4c->mbuf, MP3BUFSIZE);
	if(i > 0)
		AVI_write_audio(divx4c->avi, divx4c->mbuf, i);
	AVI_close(divx4c->avi);
	free(divx4c);
	free(fct);
}

void divx4file_core_nullvideo(struct file_core_t * fct)
{
	struct divx4file_core_t * divx4c = (struct divx4file_core_t *)fct->priv;
	
	if(!divx4c)
		return;
	AVI_write_frame(divx4c->avi, divx4c->cbuf, 0, AVI_FLAG_NONE);
}

void divx4file_core_video(struct file_core_t * fct, unsigned char **buf)
{
	struct divx4file_core_t * divx4c = (struct divx4file_core_t *)fct->priv;
	
	if(!divx4c)
		return;
	memset(&divx4c->enc_frame, 0, sizeof(ENC_FRAME));
	divx4c->enc_frame.image = buf[0];
	divx4c->enc_frame.bitstream = divx4c->cbuf;
	divx4c->enc_frame.colorspace = ENC_CSP_I420;
	if(encore(divx4c->dh, ENC_OPT_ENCODE, &divx4c->enc_frame, &divx4c->enc_result) != ENC_OK) {
		fprintf(stderr, "Error encoding frame!\n");
		AVI_write_frame(divx4c->avi, divx4c->cbuf, 0, AVI_FLAG_NONE);
	} else {
		AVI_write_frame(divx4c->avi, 
				divx4c->cbuf, 
				divx4c->enc_frame.length, 
				divx4c->enc_result.is_key_frame ? 
					AVI_FLAG_KEYFRAME : AVI_FLAG_NONE);
	}
}

void divx4file_core_sync(struct file_core_t * fct)
{
	return;
}

int divx4file_core_format(char **args)
{
	return NV_YUV420;
}

void divx4file_core_usage()
{
	fprintf(stderr, "DIVX4 file core options:\n");
	fprintf(stderr, "\t-db <divx bitrate (kbps)>               [?]\n");
	fprintf(stderr, "\t-dq <quality factor: 1-fastest, 5-best> [?]\n");
	fprintf(stderr, "\t-dqmin <minimum quantiser>              [?]\n");
	fprintf(stderr, "\t-dqmax <maximum quantizer>              [?]\n");
	fprintf(stderr, "\t-dk <maximum key interval>              [?]\n");
	fprintf(stderr, "\t-drcp <rate control period>             [?2000]\n");
	fprintf(stderr, "\t-drcrp <rate control reaction period>   [?10]\n");
	fprintf(stderr, "\t-drcrr <rate control reaction ratio>    [?20]\n");
	fprintf(stderr, "\t-deinterlace <deinterlace>              [off]\n");
	fprintf(stderr, "\t-o <output filename> [NONE]\n");
}

struct file_core_t * divx4file_core_init(char ** args, ntime_t vrate)
{
	struct file_core_t * fct;
	struct divx4file_core_t * divx4c;
	double f;
 	int arate = -1;
 	int abits = -1;
 	int achans = -1;
 	int vwidth = -1;
 	int vheight = -1;
 	char *fname = NULL;
 	ENC_PARAM enc_param;
 	NV_ARGS;
 	
 	fct =  malloc(sizeof(struct file_core_t));
 	if(!fct) {
 		perror("divx4file core init - could no malloc new cap core");
 		return NULL;
 	}
 	divx4c = malloc(sizeof(struct divx4file_core_t));
 	if(!divx4c) {
 		perror("divx4file core init - could no malloc new divx4 core");
 		free(fct);
 		return NULL;
 	}
 	memcpy(fct, &divx4file_core, sizeof(struct file_core_t));
 	fct->priv = (void *)divx4c;
 	memset(divx4c, 0, sizeof(struct divx4file_core_t));
 	
	memset(&enc_param, 0, sizeof(ENC_PARAM));
	NV_SCAN_ARGS {
		if(strcmp(NV_THIS_ARG, "-r") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			arate = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-b") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			abits = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-s") == 0) {
			achans = 2;
		} else if(strcmp(NV_THIS_ARG, "-m") == 0) {
			achans = 1;
		} else if(strcmp(NV_THIS_ARG, "-w") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			vwidth = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-h") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			vheight = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-db") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			enc_param.bitrate = atoi(NV_THIS_ARG);
			if(enc_param.bitrate <= 0)
				goto usage;
			enc_param.bitrate *= 1024;
		} else if(strcmp(NV_THIS_ARG, "-dq") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			enc_param.quality = atoi(NV_THIS_ARG);
			if((enc_param.quality < 1) || (enc_param.quality > 5))
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-dqmin") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			enc_param.min_quantizer = atoi(NV_THIS_ARG);
			if(enc_param.min_quantizer <= 0)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-dqmax") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			enc_param.max_quantizer = atoi(NV_THIS_ARG);
			if(enc_param.max_quantizer <= 0)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-dk") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			enc_param.max_key_interval = atoi(NV_THIS_ARG);
			if(enc_param.max_key_interval <= 0)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-drcp") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			enc_param.rc_period = atoi(NV_THIS_ARG);
			if(enc_param.rc_period <= 0)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-drcrp") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			enc_param.rc_reaction_period = atoi(NV_THIS_ARG);
			if(enc_param.rc_reaction_period <= 0)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-drcrr") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			enc_param.rc_reaction_ratio = atoi(NV_THIS_ARG);
			if(enc_param.rc_reaction_ratio <= 0)
				goto usage;
		} else if(strcmp(NV_THIS_ARG, "-deinterlace") == 0) {
			enc_param.deinterlace = 1;
		} else if(strcmp(NV_THIS_ARG, "-o") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			fname = NV_THIS_ARG;
		}
	}
	divx4c->ashift = (achans == 1) ? 1 : 2;

	if(!fname)
		goto usage;
	fprintf(stderr, "divx4file core init - file: %s (%dx%d)\n", fname, vwidth, vheight);
	
	divx4c->avi = AVI_open_output_file(fname);
	if(!divx4c->avi) {
		perror("divx4file core init - could not open file");
		goto error0;
	}
	f = 1e9 / (double)vrate;
	fprintf(stderr, "file core init - frame rate is %f\n", f);

	enc_param.x_dim = vwidth;
	enc_param.y_dim = vheight;
	enc_param.framerate = f;
	if(encore(NULL, ENC_OPT_INIT, &enc_param, NULL) != ENC_OK) {
		fprintf(stderr, "file core init - error initing encoder\n");
		goto error0;
	}
	divx4c->dh = enc_param.handle;
	if(!divx4c->dh)
		goto error1;
	
	fprintf(stderr, "Got DivX bitrate: %d\n", enc_param.bitrate);
	
	divx4c->cbuf = malloc(vwidth * vheight * 6);
	if(!divx4c->cbuf) {
		perror("file core init - could not allocate compression buffer");
		goto error2;
	}
	
	AVI_set_video(divx4c->avi, vwidth, vheight, f, "DIV5");

	lame_init(&divx4c->lgf);
	divx4c->lgf.num_channels = achans;
	divx4c->lgf.in_samplerate = arate;
	lame_init_params(&divx4c->lgf);
	AVI_set_audio(divx4c->avi, achans, arate, abits, 0x55, (divx4c->lgf.brate * 1000) / 8);
	
	return (void *)fct;

error2:	
	encore(divx4c->dh, ENC_OPT_RELEASE, NULL, NULL);
error1:
	AVI_close(divx4c->avi);
error0:
	free(fct);
	free(divx4c);
	return NULL;

usage:
	fprintf(stderr, "divx4file core usage:\n");
	fprintf(stderr, "Error parsing options.  Please run without any command line options for\n");
	fprintf(stderr, "detailed usage information.\n");
	free(fct);
	free(divx4c);
	return NULL;
}

struct file_core_t divx4file_core = {
	file_core_init: divx4file_core_init,
	file_core_close: divx4file_core_close,
	file_core_nullvideo: divx4file_core_nullvideo,
	file_core_fastvideo: divx4file_core_video,
	file_core_video: divx4file_core_video,
	file_core_audio: divx4file_core_audio,
	file_core_sync: divx4file_core_sync,
	file_core_format: divx4file_core_format,
	file_core_usage: divx4file_core_usage,
	priv: NULL
};
