/* 
   rtefile_core.c (C) Justin Schoeman 2001 (justin@suntiger.ee.up.ac.za)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
*/
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "rte.h"
#include "NVrec.h"

struct rtefile_core_t {
	rte_context * context;
	double vtimestep;
	double atimestep;
	void * vdata;
	unsigned char * adata;
	double atime;
	double vtime;
	int stereo;
	unsigned char abak[1632 * 2 * 2];
	int abfill;
};

void rtefile_core_usage()
{
	fprintf(stderr, "rte file (mpeg-1) core usage:\n");
	fprintf(stderr, "\t-o <output filename>                  [NONE]\n");
	fprintf(stderr, "\t-vb <kbits per second for video>      [2300]\n");
	fprintf(stderr, "\t-ab <kbits per second for audio>      [80]\n");
}

struct file_core_t rtefile_core;

struct file_core_t * rtefile_core_init(char ** args, ntime_t vrate)
{
	struct file_core_t * fct;
	struct rtefile_core_t * rtec;
 	int arate = -1;
 	int abits = -1;
 	int width = -1;
 	int height = -1;
 	int vb = -1;
 	int ab = -1;
 	char *fname = NULL;
 	NV_ARGS;
 	
 	fct = (struct file_core_t *)malloc(sizeof(struct file_core_t));
 	if(!fct) {
 		perror("rtefile core init - could no malloc new file core");
 		return NULL;
 	}
 	rtec = (struct rtefile_core_t *)malloc(sizeof(struct rtefile_core_t));
 	if(!rtec) {
 		perror("rtefile core init - could no malloc new rte core");
 		free(fct);
 		return NULL;
 	}
 	memcpy(fct, &rtefile_core, sizeof(struct file_core_t));
 	fct->priv = (void *)rtec;
 	
	NV_SCAN_ARGS{
		if(strcmp(NV_THIS_ARG, "-r") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			arate = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-b") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			abits = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-s") == 0) {
			rtec->stereo = 1;
		} else if(strcmp(NV_THIS_ARG, "-m") == 0) {
			rtec->stereo = 0;
		} else if(strcmp(NV_THIS_ARG, "-w") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			width = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-h") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			height = atoi(NV_THIS_ARG);
		} else if(strcmp(NV_THIS_ARG, "-vb") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			vb = atoi(NV_THIS_ARG);
			vb *= 1000;
		} else if(strcmp(NV_THIS_ARG, "-ab") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			ab = atoi(NV_THIS_ARG);
			ab *= 1000;
			if (ab < 80000)
				fprintf(stderr, "rtefile core init - setting audio bit rate too low.. recording may stop unexpectedly\n");
		} else if(strcmp(NV_THIS_ARG, "-o") == 0) {
			if(!NV_NEXT_ARG)
				goto usage;
			fname = NV_THIS_ARG;
		}
	}
	if((!fname) || (width&15) || (height&15))
		goto usage;
	fprintf(stderr, "file core init - file: %s\n", fname);

	if(!rte_init()) {
		fprintf(stderr, "file core init - cannot run rte\n");
		goto error0;
	} 
	
	rtec->context = rte_context_new(width, height, RTE_YUV420, (void*)0xdeadbeef);
	if(!rtec->context) {
		fprintf(stderr, "file core init - could not allocate rte context\n");
		goto error0;
	} 
	rte_set_verbosity(rtec->context, 0);
	rte_set_mode(rtec->context, RTE_AUDIO_AND_VIDEO);
	rte_set_audio_parameters(rtec->context, arate, 
		(rtec->stereo) ? RTE_AUDIO_MODE_STEREO : RTE_AUDIO_MODE_MONO, 
		(ab > 0) ? ab : rtec->context->output_audio_bits);
	rte_set_video_parameters(rtec->context, RTE_YUV420, width, height, 
		(vrate == 40000000ULL) ? RTE_RATE_3 : RTE_RATE_4, 
		(vb > 0) ? vb : rtec->context->output_video_bits);
	rte_set_input(rtec->context, RTE_AUDIO, RTE_PUSH, FALSE, 
		NULL, NULL, NULL);
	rte_set_input(rtec->context, RTE_VIDEO, RTE_PUSH, FALSE, 
		NULL, NULL, NULL);
	rte_set_output(rtec->context, NULL, fname);
	if (!rte_init_context(rtec->context)) {
		fprintf(stderr, "file core init - could not init rte context\n");
		goto error1;
	}
	rtec->vtimestep = ((double)vrate) / 1e9;
	rtec->atimestep = ((double)1152) / ((double)arate);
	rtec->vtime = 0;
	rtec->atime = 0;
	rtec->abfill = 0;
	if (!rte_start_encoding(rtec->context)) {
		fprintf(stderr, "file core init - could not start rte encoding\n");
		goto error1;
	}
	rtec->vdata = rte_push_video_data(rtec->context, NULL, 0);
	rtec->adata = rte_push_audio_data(rtec->context, NULL, 0);
	return fct;

error1:
	rte_context_destroy(rtec->context);
error0:
	free(fct);
	free(rtec);
	return NULL;
usage:
	fprintf(stderr, "Error parsing options.  Please run without any command line options for\n");
	fprintf(stderr, "detailed usage information.\n");
	free(fct);
	free(rtec);
	return NULL;
}

void rtefile_core_close(struct file_core_t * fct)
{
	struct rtefile_core_t * rtec = (struct rtefile_core_t *)fct->priv;
	
	if(!rtec)
		return;
	rte_stop(rtec->context);
	rte_context_destroy(rtec->context);
	free(rtec);
	free(fct);
}

void rtefile_core_nullvideo(struct file_core_t * fct)
{
	struct rtefile_core_t * rtec = (struct rtefile_core_t *)fct->priv;
	if(!rtec)
		return;
	rtec->vtime += rtec->vtimestep;
}

void rtefile_core_video(struct file_core_t * fct, unsigned char **buf)
{
	struct rtefile_core_t * rtec = (struct rtefile_core_t *)fct->priv;
	
	if(!rtec)
		return;
		
	memcpy(rtec->vdata, buf[0], rtec->context->video_bytes);
	rtec->vdata = rte_push_video_data(rtec->context, rtec->vdata, rtec->vtime);
	if(!rtec->vdata)
		fprintf(stderr, "Uh-Oh.. encoding error, about to die...\n");
	rtec->vtime += rtec->vtimestep;
}

void rtefile_core_audio(struct file_core_t * fct, unsigned char *buf, int n)
{
	struct rtefile_core_t * rtec = (struct rtefile_core_t *)fct->priv;
	int c;

	if(!rtec)
		return;
		
	while((rtec->abfill + n) >= rtec->context->audio_bytes) {
		if(rtec->abfill > 0)
			memcpy(rtec->adata, rtec->abak, rtec->abfill);
		c = rtec->context->audio_bytes - rtec->abfill;
		memcpy(rtec->adata + rtec->abfill, buf, c);
		memcpy(rtec->abak, rtec->adata + (2304 << rtec->stereo), 
			(960 << rtec->stereo));
		rtec->abfill = 960 << rtec->stereo;
		n -= c;
		buf += c;
		rtec->adata = rte_push_audio_data(rtec->context, 
			rtec->adata, rtec->atime);
		if(!rtec->adata)
			fprintf(stderr, "Uh-Oh.. encoding error, about to die...\n");
		rtec->atime += rtec->atimestep;
	}
	if(n) {
		memcpy(rtec->abak + rtec->abfill, buf, n);
		rtec->abfill += n;
	}
}

void rtefile_core_sync(struct file_core_t * fct)
{
	return;
}

inline int rtefile_core_format(char **args)
{
	return NV_YUV420;
}

struct file_core_t rtefile_core = {
	file_core_init: rtefile_core_init,
	file_core_close: rtefile_core_close,
	file_core_nullvideo: rtefile_core_nullvideo,
	file_core_fastvideo: rtefile_core_video,
	file_core_video: rtefile_core_video,
	file_core_audio: rtefile_core_audio,
	file_core_sync: rtefile_core_sync,
	file_core_format: rtefile_core_format,
	file_core_usage: rtefile_core_usage,
	priv: NULL
};
